# -*- coding: utf-8 -*-

from bts.api import BTS
import json
import os


class LightWallet(object):
    def __init__(self):
        config_file = os.getenv("HOME") + "/.python-bts/bts_client.json"
        fd_config = open(config_file)
        config_bts = json.load(fd_config)["client_default"]
        fd_config.close()

        self.bts_client = BTS(config_bts["user"], config_bts["password"],
                              config_bts["host"], config_bts["port"])
        self.pusher = None
        self.wallets = {}
        client_info = self.bts_client.get_info()
        self.height = int(client_info["blockchain_head_block_num"])

    def myPublish(self, topic, event):
        if self.pusher:
            self.pusher.emit(topic, event, namespace="")

    def get_wallet(self, address):
        if address not in self.wallets:
            wallet = self.init_wallet(address)
            if not wallet:
                return None
            self.wallets[address] = wallet
        return self.wallets[address]

    def init_wallet(self, address):
        wallet = {}
        wallet["address"] = address
        wallet["account"] = self.get_account(address)
        wallet["balance"] = self.get_balance(address)
        wallet["trx"] = self.get_trx(address)
        wallet["need_update_balance"] = False
        return wallet

    def get_account(self, address):
        account_info = self.bts_client.get_account_info(address)
        if account_info:
            return account_info["name"]
        else:
            return ""

    def get_balance(self, address):
        balances = self.bts_client.get_address_balances(address)
        if balances:
            return balances
        else:
            return {}

    def get_trx(self, address):
        trxs = []
        trxs_info = self.bts_client.request(
            "blockchain_list_address_transactions", [address]
        ).json()["result"]
        for trx_id in trxs_info:
            _trx = trxs_info[trx_id]["trx"]
            _op = _trx["trx"]['operations'][0]
            if _op["type"] == 'update_feed_op_type' or \
                    _op["type"] == 'register_account_op_type' or \
                    _op["type"] == 'update_account_op_type':
                continue
            trx = {}
            owner_from = ""
            owner_to = ""
            trx["memo"] = None
            for _op in _trx["trx"]['operations']:
                if _op["type"] == "withdraw_op_type":
                    balance_id = _op["data"]["balance_id"]
                    balance_info = self.bts_client.request(
                        "blockchain_get_balance", [balance_id]
                    ).json()["result"]
                    owner_from = balance_info["condition"]["data"]["owner"]
                    continue
                if _op["type"] == "deposit_op_type":
                    trx["asset"] = self.bts_client.get_asset_symbol(
                        _op["data"]["condition"]["asset_id"])
                    trx["amount"] = float(_op["data"]["amount"]) / \
                        self.bts_client.get_asset_precision(trx["asset"])
                    if "owner" in _op["data"]["condition"]["data"]:
                        owner_to = _op["data"]["condition"]["data"]["owner"]
                    if "memo" in _op["data"]["condition"]["data"]:
                        trx["memo"] = _op["data"]["condition"]["data"]["memo"]
                    continue
            trx["trx_id"] = trx_id[:8]
            trx["block_num"] = _trx["chain_location"]["block_num"]
            trx["timestamp"] = self.bts_client.get_block_timestamp(
                trx["block_num"])
            trx["fee_asset"] = \
                self.bts_client.get_asset_symbol(_trx["fees_paid"][0][0])
            trx["fee_amount"] = float(_trx["fees_paid"][0][1]) / \
                self.bts_client.get_asset_precision(_trx["fees_paid"][0][0])

            if owner_from == address:
                trx["account"] = self.get_account(owner_to)
                trx["type"] = "send"
            elif owner_to == address:
                trx["account"] = self.get_account(owner_from)
                trx["type"] = "receive"
            trxs.append(trx)
        return sorted(trxs, key=lambda item:item["block_num"])

    #def init_transaction(self):
    #    self.transaction = {}
    #    trxs = self.bts_client.get_transaction_history(limit=-50)
    #    for account in self.get_all_account():
    #        self.transaction[account] = \
    #            self.bts_client.format_transaction_history(account, trxs)

    #def update_transaction(self, start, end):
    #    trxs = self.bts_client.get_transaction_history(
    #        limit=-500, start=start, end=end)
    #    if not trxs:
    #        return
    #    for account in self.get_all_account():
    #        if account not in self.transaction:
    #            self.transaction[account] = []
    #        format_trxs = \
    #            self.bts_client.format_transaction_history(account, trxs)
    #        if not format_trxs:
    #            continue
    #        self.transaction[account].extend(format_trxs)
    #        self.myPublish("user."+account+".transaction", format_trxs)

    def update_trx(self, trxs, virtual_trxs):
        for _trx in trxs:
            _op = _trx[1]["trx"]['operations'][0]
            if _op["type"] == 'update_feed_op_type' or \
                    _op["type"] == 'register_account_op_type' or \
                    _op["type"] == 'update_account_op_type':
                continue
            trx = {}
            owner_from = ""
            owner_to = ""
            trx["memo"] = None
            for _op in _trx[1]["trx"]['operations']:
                if _op["type"] == "withdraw_op_type":
                    balance_id = _op["data"]["balance_id"]
                    balance_info = self.bts_client.request(
                        "blockchain_get_balance", [balance_id]
                    ).json()["result"]
                    owner_from = balance_info["condition"]["data"]["owner"]
                    continue
                if _op["type"] == "deposit_op_type":
                    trx["asset"] = self.bts_client.get_asset_symbol(
                        _op["data"]["condition"]["asset_id"])
                    trx["amount"] = float(_op["data"]["amount"]) / \
                        self.bts_client.get_asset_precision(trx["asset"])
                    if "owner" in _op["data"]["condition"]["data"]:
                        owner_to = _op["data"]["condition"]["data"]["owner"]
                    if "memo" in _op["data"]["condition"]["data"]:
                        trx["memo"] = _op["data"]["condition"]["data"]["memo"]
                    continue
            if owner_from not in self.wallets and owner_to not in self.wallets:
                continue

            trx["trx_id"] = _trx[0][:8]
            trx["block_num"] = _trx[1]["chain_location"]["block_num"]
            trx["timestamp"] = self.bts_client.get_block_timestamp(
                trx["block_num"])
            trx["fee_asset"] = \
                self.bts_client.get_asset_symbol(_trx[1]["fees_paid"][0][0])
            trx["fee_amount"] = float(_trx[1]["fees_paid"][0][1]) / \
                self.bts_client.get_asset_precision(_trx[1]["fees_paid"][0][0])

            if owner_from in self.wallets:
                trx_tmp = trx.copy()
                trx_tmp["account"] = self.get_account(owner_to)
                trx_tmp["type"] = "send"
                self.myPublish("user."+owner_from+".transaction", trx_tmp)
                self.wallets[owner_from]["trx"].append(trx_tmp)
                self.wallets[owner_from]["need_update_balance"] = True

            if owner_to in self.wallets and owner_to != owner_from:
                trx_tmp = trx.copy()
                trx_tmp["account"] = self.get_account(owner_from)
                trx_tmp["type"] = "receive"
                self.myPublish("user."+owner_to+".transaction", trx_tmp)
                self.wallets[owner_to]["trx"].append(trx_tmp)
                self.wallets[owner_to]["need_update_balance"] = True

    def update_balance(self):
        for address in self.wallets:
            if self.wallets[address]["need_update_balance"]:
                self.wallets[address]["balance"] = self.get_balance(address)
                self.wallets[address]["need_update_balance"] = False
                self.myPublish(
                    "user."+address+".balance",
                    self.wallets[address]["balance"])

    def transfer(self, trx):
        pass

    def execute(self):
        client_info = self.bts_client.get_info()
        height_now = int(client_info["blockchain_head_block_num"])
        if(height_now <= self.height):
            return

        time_stamp = client_info["blockchain_head_block_timestamp"]
        self.myPublish(u'bts.blockchain.info',
                       {"height": height_now, "time_stamp": time_stamp})
        while self.height < height_now:
            self.height += 1
            trxs = self.bts_client.get_block_transactions(
                self.height)
            #virtual_trxs = self.market.get_order_deal_rec(self.height)
            virtual_trxs = []
            self.update_trx(trxs, virtual_trxs)
        self.update_balance()



from gevent import monkey
monkey.patch_all()

#import datetime
import time
from threading import Thread
#from flask import Flask, render_template, session, request
from flask import Flask, render_template, request, redirect, url_for
from flask.ext.babel import gettext#, ngettext
#from flask import send_from_directory
from flask import flash
from flask.ext.socketio import SocketIO, emit
from flask.ext.babel import Babel
from bts_wallet import BTSWallet
from bts_orderbook import BTSOrderBook
from pprint import pprint

app = Flask(__name__, static_folder='static', static_url_path='')
app.debug = True
app.config['SECRET_KEY'] = 'secret!'
socketio = SocketIO(app)
babel = Babel(app)
bts_orderbook = BTSOrderBook()
bts_orderbook.pusher = socketio
contacts = ["test-alice", "test-bob", "rou.baozi"]
protect_account = ["pls.btsbots", "alice.btsbots", "bob.btsbots"]
bts_wallet = BTSWallet(bts_orderbook.bts_client, protect_account)
bts_wallet.pusher = socketio

@babel.localeselector
def get_locale():
    lang = ["en", "zh"]
    return request.accept_languages.best_match(lang)


def background_thread():
    """Example of how to send server generated events to clients."""
    while True:
        time.sleep(10)
        bts_wallet.execute()
        bts_orderbook.execute()


def get_wallet_account():
    _account = request.cookies.get('account')
    account_list = bts_wallet.get_all_account()
    if _account and _account in account_list:
        return _account
    else:
        return None


@app.template_filter()
def datetimefilter(value, format='%Y/%m/%d %H:%M'):
    """convert a datetime to a different format."""
    return value.strftime(format)

app.jinja_env.filters['datetimefilter'] = datetimefilter


@app.route('/')
def index():
    account = get_wallet_account()
    if not account:
        return redirect(url_for('waccount'))
    current_height = bts_orderbook.height
    balance = bts_wallet.get_balance(account)
    transaction = bts_wallet.get_transaction(account)
    return render_template(
        'index.html', title=gettext(u"Wallet"), current_height=current_height,
        account=account, balance=balance, transaction=transaction
    )


@app.route('/transfer', methods=['Get', 'POST'])
def transfer():
    account = get_wallet_account()
    if not account:
        return redirect(url_for('waccount'))
    balance = bts_wallet.get_balance(account)
    _contacts = contacts
    msg = ""
    if request.method == "POST":
        print(request.form)
        amount = request.form["amount"]
        if amount:
            amount = float(amount)
        else:
            amount = 0.0
        asset = request.form["asset"]
        receiver = request.form["receiver"]
        memo = request.form["memo"]
        if amount <= 0.0 or amount > balance[asset]:
            msg = "Error: amount invalid %s %s" % (amount, balance[asset])
        elif receiver not in _contacts:
            msg = "Error: account %s invalid" % receiver
        else:
            result = bts_wallet.transfer([amount, asset, account, receiver, memo])
            if "Error" in result:
                msg = "Error: %s" % result["error"]["message"]
            else:
                msg = "Success: transfer %s %s from %s to %s %s" % (
                    amount, asset, account, receiver, memo)

    return render_template(
        'transfer.html', title=gettext(u"Transfer"), contacts=_contacts,
        account=account, balance=balance, msg=msg
    )


@app.route('/waccount')
def waccount():
    account_list = bts_wallet.get_all_account()
    return render_template(
        'waccount.html', title="choose account",
        account_list=account_list)


@app.route('/contact')
def contact():
    #account_list = bts_wallet.get_all_account()
    _contacts = contacts
    return render_template(
        'contact.html', title="contact list",
        contacts=_contacts)


@app.route('/market')
def market():
    current_height = bts_orderbook.height
    order_book = bts_orderbook.order_book["CNY_BTS"]
    deal_trx = bts_orderbook.deal_trx["CNY_BTS"]
    place_trx = bts_orderbook.place_trx["CNY_BTS"]
    return render_template(
        'market.html', title=gettext("Market"), current_height=current_height,
        order_book=order_book, deal_trx=deal_trx, place_trx=place_trx
    )


@socketio.on('connect', namespace='')
def test_connect():
    emit('my response', {'data': 'Connected', 'count': 0})


def main():
    thread = Thread(target=background_thread)
    thread.start()

    socketio.run(
        app, use_reloader=False, heartbeat_interval=10, heartbeat_timeout=15)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#from bts.api import BTS
#import json
#import os


class BTSWallet(object):
    def __init__(self, client, protect_account=[]):
        self.protect_account = protect_account
        self.bts_client = client
        self.pusher = None
        self.balance = {}
        client_info = self.bts_client.get_info()
        self.height = int(client_info["blockchain_head_block_num"])
        self.accounts = None
        self.update_accounts()
        self.update_balance()
        self.trx = None
        self.update_trx()
        self.init_transaction()

    def myPublish(self, topic, event):
        if self.pusher:
            self.pusher.emit(topic, event, namespace="")

    def update_accounts(self):
        accounts = self.bts_client.list_accounts()
        for account in list(accounts):
            if account in self.protect_account:
                accounts.remove(account)
        if accounts != self.accounts:
            self.myPublish("account_list", accounts)
            self.accounts = accounts

    def get_all_account(self):
        return self.accounts

    def get_balance(self, account):
        if account not in self.balance:
            return {}
        else:
            return self.balance[account]

    def init_transaction(self):
        self.transaction = {}
        trxs = self.bts_client.get_transaction_history(limit=-50)
        for account in self.get_all_account():
            self.transaction[account] = \
                self.bts_client.format_transaction_history(account, trxs)

    def update_transaction(self, start, end):
        trxs = self.bts_client.get_transaction_history(
            limit=-500, start=start, end=end)
        if not trxs:
            return
        for account in self.get_all_account():
            if account not in self.transaction:
                self.transaction[account] = []
            format_trxs = \
                self.bts_client.format_transaction_history(account, trxs)
            if not format_trxs:
                continue
            self.transaction[account].extend(format_trxs)
            self.myPublish("user."+account+".transaction", format_trxs)

    def get_transaction(self, account):
        if account in self.transaction:
            return self.transaction[account]
        else:
            return []

    def update_balance(self):
        balance = self.bts_client.get_balance()
        for _account in self.get_all_account():
            if _account not in self.balance:
                self.balance[_account] = {"BTS":0}
            if _account not in balance:
                balance[_account] = {"BTS":0}
            if balance[_account] != self.balance[_account]:
                self.myPublish("user."+_account+".balance", balance[_account])
                self.balance[_account] = balance[_account]

    def transfer(self, trx):
        return self.bts_client.transfer(trx)

    def update_trx(self, start=0, end=-1):
        pass
        #trx = self.bts_client.get_trx()
        #if trx is not None:
        #    self.trx.extend(trx)
        #    self.myPublish("trx", trx)

    def execute(self):
        client_info = self.bts_client.get_info()
        height_now = int(client_info["blockchain_head_block_num"])
        if(height_now <= self.height):
            return

        self.update_balance()
        self.update_accounts()
        self.update_trx()
        self.update_transaction(self.height+1, height_now)
        self.height = height_now



#from gevent import monkey
#monkey.patch_all()

#import datetime
import time
from threading import Thread
#from flask import Flask, render_template, session, request
from flask import Flask, render_template, request, redirect, url_for
from flask.ext.babel import gettext
#from flask import send_from_directory
#from flask import flash
from flask.ext.socketio import SocketIO, emit
from flask.ext.babel import Babel
#from pprint import pprint

# from pprint import pprint
from bts.ws.base_protocol import BaseProtocol
import datetime

try:
    import asyncio
except ImportError:
    import trollius as asyncio

app = Flask(__name__, static_folder='static', static_url_path='')
app.debug = True
app.config['SECRET_KEY'] = 'secret!'
socketio = SocketIO(app)
babel = Babel(app)
contacts = ["test-alice", "test-bob", "rou.baozi"]

def id_to_int(id):
    return int(id.split('.')[-1])

def next_id(id):
    id_array = id.split('.')
    id_next = "%s.%s.%d" % (id_array[0], id_array[1], int(id_array[2])+1)
    return id_next

class ChainMonitor(BaseProtocol):
    op_id_last = None
    object_info = {}
    last_fill_order = None
    lock = asyncio.Lock()

    @asyncio.coroutine
    def get_object_info(self, id):
        if id not in self.object_info:
            response = yield from self.rpc([self.database_api, "get_objects", [[id]]])
            self.object_info[id] = response[0]
        return self.object_info[id]

    @asyncio.coroutine
    def handle_op_transfer(self, notify):
        _info = notify["op"][1]
        _transfer_info = {}
        response = yield from self.get_object_info(_info["from"])
        _transfer_info["from"] = response["name"]
        response = yield from self.get_object_info(_info["to"])
        _transfer_info["to"] = response["name"]
        _asset_info = yield from self.get_object_info(_info["amount"]["asset_id"])
        _transfer_info["symbol"] = _asset_info["symbol"]
        _transfer_info["amount"] = float(_info["amount"]["amount"])/10**_asset_info["precision"]
        print("[33m[%s] %s sent %s %s to %s[m" % (
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), _transfer_info["from"],
            _transfer_info["amount"], _transfer_info["symbol"], _transfer_info["to"]))

    @asyncio.coroutine
    def handle_op_borrow(self, notify):
        _info = notify["op"][1]
        _borrow_info = {}
        response = yield from self.get_object_info(_info["funding_account"])
        _borrow_info["account"] = response["name"]
        for _type in ["delta_collateral", "delta_debt"]:
            _balance = _info[_type]
            _asset_info = yield from self.get_object_info(_balance["asset_id"])
            _balance["symbol"] = _asset_info["symbol"]
            _balance["amount"] = float(_balance["amount"])/10**_asset_info["precision"]
            _borrow_info[_type] = _balance
        print("[31m[%s] %s adjust collateral by %s %s, debt by %s %s[m" % (
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), _borrow_info["account"],
            _borrow_info["delta_collateral"]["amount"], _borrow_info["delta_collateral"]["symbol"],
            _borrow_info["delta_debt"]["amount"], _borrow_info["delta_debt"]["symbol"]))

    @asyncio.coroutine
    def handle_op_fill(self, notify):
        _info = notify["op"][1]
        if self.last_fill_order is None:
            self.last_fill_order = _info
            return
        if _info["pays"] != self.last_fill_order["receives"] or \
          _info["receives"] != self.last_fill_order["pays"]:
            print("error1: %s" % self.last_fill_order)
            print("error2: %s" % _info)
            self.last_fill_order = _info
            return
        _trade_info = {}
        _order_info = self.last_fill_order
        self.last_fill_order = None
        response = yield from self.get_object_info(_order_info["account_id"])
        _trade_info["account1"] = response["name"]
        response = yield from self.get_object_info(_info["account_id"])
        _trade_info["account2"] = response["name"]
        for _type in ["pays", "receives"]:
            _balance = _order_info[_type]
            _asset_info = yield from self.get_object_info(_balance["asset_id"])
            _balance["symbol"] = _asset_info["symbol"]
            _balance["amount"] = float(_balance["amount"])/10**_asset_info["precision"]
            _trade_info[_type] = _balance

        print("[32m[%s] %s bought %s %s with %s %s from %s[m" % (
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), _trade_info["account1"],
            _trade_info["receives"]["amount"], _trade_info["receives"]["symbol"],
            _trade_info["pays"]["amount"], _trade_info["pays"]["symbol"],
            _trade_info["account2"]))

    handles = {
        0: handle_op_transfer,
        3: handle_op_borrow,
        4: handle_op_fill}

    @asyncio.coroutine
    def handle_operation(self, notify):
        _op_type = notify["op"][0]
        print(notify["id"])
        if _op_type in self.handles:
            yield from self.handles[_op_type](self, notify)

    def onOperation(self, notify):
        asyncio.async(self._onOperation(notify))

    @asyncio.coroutine
    def _onOperation(self, notify):
        yield from self.lock
        op_id_cur = notify["id"]
        if self.op_id_last is None:
            self.op_id_last = op_id_cur
        op_id_last = self.op_id_last
        if id_to_int(op_id_last) > id_to_int(op_id_cur):
            self.lock.release()
            return
        while id_to_int(op_id_last) < id_to_int(op_id_cur):
            response = yield from self.rpc([self.database_api, "get_objects", [[op_id_last]]])
            _notify = response[0]
            yield from self.handle_operation(_notify)
            op_id_last = next_id(op_id_last)
        yield from self.handle_operation(notify)
        self.op_id_last = next_id(op_id_last)
        self.lock.release()

    @asyncio.coroutine
    def onOpen(self):
        op_id_last = self.op_id_last
        yield from super().onOpen()
        self.subscribe("1.11.", self.onOperation)


@babel.localeselector
def get_locale():
    lang = ["en", "zh"]
    return request.accept_languages.best_match(lang)


def background_thread():
    """Example of how to send server generated events to clients."""
    time.sleep(10)


@app.template_filter()
def datetimefilter(value, format='%Y/%m/%d %H:%M'):
    """convert a datetime to a different format."""
    return value.strftime(format)

app.jinja_env.filters['datetimefilter'] = datetimefilter


@app.route('/')
def index():
    address = request.cookies.get('address')
    return render_template(
        'index.html', title=gettext(u"explorer")
    )

@socketio.on('connect', namespace='')
def test_connect():
    emit('my response', {'data': 'Connected', 'count': 0})


def main():
    thread = Thread(target=background_thread)
    thread.start()

    socketio.run(
        app, use_reloader=False)

if __name__ == '__main__':
    main()
